@@ -5,7 +5,7 @@ Author: Zentetsu
 
 ----
 
-Last Modified: Fri Nov 26 2021
+Last Modified: Mon Mar 06 2023
 Modified By: Zentetsu
 
 ----
@@ -39,6 +39,9 @@ HISTORY:
 2021-11-24	Zen	SharedMemory init size fixed
 2021-11-25	Zen	Changing Server behavior
 2021-11-26	Zen	Fix int and dict encoding
+2021-11-27	Zen	Fix getValue behavior
+2021-11-27	Zen	Adding semaphore
+2023-03-06	Zen	Correcting mutex behavior + Fixing mutex bug
 '''
 
 from abc import ABC, abstractmethod
@@ -52,6 +55,8 @@ import sys
 import os
 
 _SHM_NAME_PREFIX = '/psm_'
+_SEM_NAME_PREFIX = '/sem_'
+
 _MODE = 0o600
 _FLAG = posix_ipc.O_CREX | os.O_RDWR
 
@@ -81,13 +86,16 @@ class SharedMemory:
             path (str, optional): path to load JSON file and sharing data inside. Defaults to None.
             size (int, optional): size of the shared space or str value. Defaults to 10.
             client (bool, optional): will creat a client or server instance
-            timeout (int, optional): mutex timeout. Defaults to 1.
 
         Raises:
             SMMultiInputError: raise an error when value and path are both at None or initialized
         """
         self.__log = log
 
+        if self.__log is not None:
+            f = open(os.devnull, 'w')
+            sys.stdout = f
+
         if client and (value is None and path is None or value is not None and path is not None):
             if self.__log is not None:
                 self.__writeLog(1, "Conflict between value and json path intialization.")
@@ -104,7 +112,8 @@ class SharedMemory:
             self.__value = value
             self.__size = size
 
-        self.__name = _SHM_NAME_PREFIX + name
+        self.__name_memory = _SHM_NAME_PREFIX + name
+        self.__name_semaphore = _SEM_NAME_PREFIX + name
         self.__type = type(self.__value)
         self.__client = client
 
@@ -142,6 +151,8 @@ class SharedMemory:
 
             return
 
+        self.__semaphore.acquire()
+
         if self.__mapfile is not None:
             self.__mapfile.seek(0)
 
@@ -154,14 +165,18 @@ class SharedMemory:
 
         if self.__memory is not None:
             try:
-                posix_ipc.unlink_shared_memory(self.__name)
+                posix_ipc.unlink_shared_memory(self.__name_memory)
+
+                self.__semaphore.release()
+                self.__semaphore.unlink()
             except:
                 if self.__log:
                     self.__writeLog(0, "SharedMemory already closed.")
 
             self.__memory = None
+            self.__semaphore = None
 
-    def setValue(self, value) -> bool:
+    def setValue(self, value, mutex=False) -> bool:
         """Method to set the shared value
 
         Args:
@@ -170,13 +185,15 @@ class SharedMemory:
         Returns:
             bool: return if value has been updated
         """
-        if not self.getAvailability():
+        if not self.getAvailability() or self.__semaphore is None:
             if self.__log is not None:
                 self.__writeLog(1, "Shared Memory space doesn't exist.")
             else:
                 print("ERROR: Shared Memory space doesn't exist.")
 
-            return False
+            return None
+        elif not mutex:
+            self.__semaphore.acquire()
 
         if not self.__checkValue(type(value)):
             raise SMTypeError()
@@ -188,28 +205,35 @@ class SharedMemory:
         self.__mapfile.seek(0)
         self.__mapfile.write(_packed)
 
+        if not mutex:
+            self.__semaphore.release()
+
         return True
 
-    def getValue(self):
+    def getValue(self, mutex=False):
         """Method to return the shared value
 
         Returns:
             [type]: return data from the shared space
         """
-        if not self.getAvailability():
+        if not self.getAvailability() or self.__semaphore is None:
             if self.__log is not None:
                 self.__writeLog(1, "Shared Memory space doesn't exist.")
             else:
                 print("ERROR: Shared Memory space doesn't exist.")
 
             return None
+        elif not mutex:
+            self.__semaphore.acquire()
 
-        self.__mapfile.seek(0)
-
-        _packed = self.__mapfile.read()
-        _encoded_data = list(struct.unpack('<%dI' % (len(_packed) // 4), _packed))
-        _res = len(_encoded_data) - 1 - _encoded_data[::-1].index(_END)
-        _encoded_data = _encoded_data[:_res+1]
+        try:
+            self.__mapfile.seek(0)
+            _packed = self.__mapfile.read()
+            _encoded_data = list(struct.unpack('<%dI' % (len(_packed) // 4), _packed))
+            _res = len(_encoded_data) - 1 - _encoded_data[::-1].index(_END)
+            _encoded_data = _encoded_data[:_res+1]
+        except:
+            return None
 
         if _encoded_data[0] != _BEGIN:
             raise SMEncoding("BEGIN")
@@ -230,6 +254,9 @@ class SharedMemory:
 
         self.__value == _decoded_data
 
+        if not mutex:
+            self.__semaphore.release()
+
         return _decoded_data
 
     def getType(self):
@@ -285,21 +312,25 @@ class SharedMemory:
         """
         self.__size = sys.getsizeof(self.__encoding(self.__value))
         self.__memory = None
+        self.__semaphore = None
         self.__mapfile = None
 
         if self.__client:
             try:
-                self.__memory = posix_ipc.SharedMemory(self.__name, _FLAG, _MODE)
+                self.__memory = posix_ipc.SharedMemory(self.__name_memory, _FLAG, _MODE)
+                self.__semaphore = posix_ipc.Semaphore(self.__name_semaphore, _FLAG, _MODE, initial_value=1)
                 os.ftruncate(self.__memory.fd, self.__size)
             except posix_ipc.ExistentialError:
                 if self.__value is not None:
-                    self.__memory = posix_ipc.SharedMemory(self.__name)
+                    self.__memory = posix_ipc.SharedMemory(self.__name_memory)
+                    self.__semaphore = posix_ipc.Semaphore(self.__name_semaphore)
                 else:
                     self.close()
-                    raise SMAlreadyExist(self.__name)
+                    raise SMAlreadyExist(self.__name_memory)
         else:
             try:
-                self.__memory = posix_ipc.SharedMemory(self.__name)
+                self.__memory = posix_ipc.SharedMemory(self.__name_memory)
+                self.__semaphore = posix_ipc.Semaphore(self.__name_semaphore)
                 self.__size = self.__memory.size
             except posix_ipc.ExistentialError:
                 if self.__log is not None:
@@ -582,15 +613,19 @@ class SharedMemory:
         Raises:
             TypeError: raise an error when this method is called and tha data shared type is not a dict
         """
-        if not self.getAvailability():
+        if not self.getAvailability() or self.__semaphore is None:
             if self.__log is not None:
                 self.__writeLog(1, "Shared Memory space doesn't exist.")
             else:
                 print("ERROR: Shared Memory space doesn't exist.")
 
             return None
+        else:
+            self.__semaphore.acquire()
 
-        self.__value = self.getValue()
+        self.__value = self.getValue(mutex=True)
+
+        self.__semaphore.release()
 
         if self.__type == dict:
             if type(key) is int:
@@ -612,13 +647,15 @@ class SharedMemory:
         Raises:
             TypeError: raise an error when this method is called and tha data shared type is not a dict
         """
-        if not self.getAvailability():
+        if not self.getAvailability() or self.__semaphore is None:
             if self.__log is not None:
                 self.__writeLog(1, "Shared Memory space doesn't exist.")
             else:
                 print("ERROR: Shared Memory space doesn't exist.")
 
             return None
+        else:
+            self.__semaphore.acquire()
 
         if self.__type == list:
             if self.__log is not None:
@@ -626,7 +663,7 @@ class SharedMemory:
 
             raise TypeError("Data shared type is list not dict.")
 
-        self.__value = self.getValue()
+        self.__value = self.getValue(mutex=True)
 
         if type(key) is int:
             key = str(key)
@@ -643,7 +680,9 @@ class SharedMemory:
         if sys.getsizeof(self.__value) > self.__size:
             raise SMSizeError
 
-        self.setValue(self.__value)
+        self.setValue(self.__value, mutex=True)
+
+        self.__semaphore.release()
 
     def __len__(self) -> int:
         """Method that returns the size of the shared data
@@ -654,15 +693,19 @@ class SharedMemory:
         Raises:
             TypeError: raise an error when this method is called and tha data shared type is not a dict
         """
-        if not self.getAvailability():
+        if not self.getAvailability() or self.__semaphore is None:
             if self.__log is not None:
                 self.__writeLog(1, "Shared Memory space doesn't exist.")
             else:
                 print("ERROR: Shared Memory space doesn't exist.")
 
             return None
+        else:
+            self.__semaphore.acquire()
 
-        self.__value = self.getValue()
+        self.__value = self.getValue(mutex=True)
+
+        self.__semaphore.release()
 
         return self.__value.__len__()
 
@@ -678,15 +721,19 @@ class SharedMemory:
         Raises:
             TypeError: raise an error when this method is called and tha data shared type is not a dict
         """
-        if not self.getAvailability():
+        if not self.getAvailability() or self.__semaphore is None:
             if self.__log is not None:
                 self.__writeLog(1, "Shared Memory space doesn't exist.")
             else:
                 print("ERROR: Shared Memory space doesn't exist.")
 
             return None
+        else:
+            self.__semaphore.acquire()
 
-        self.__value = self.getValue()
+        self.__value = self.getValue(mutex=True)
+
+        self.__semaphore.release()
 
         return self.__value.__contains__(key)
 
@@ -699,19 +746,23 @@ class SharedMemory:
         Raises:
             TypeError: raise an error when this method is called and tha data shared type is not a dict
         """
-        if not self.getAvailability():
+        if not self.getAvailability() or self.__semaphore is None:
             if self.__log is not None:
                 self.__writeLog(1, "Shared Memory space doesn't exist.")
             else:
                 print("ERROR: Shared Memory space doesn't exist.")
 
             return None
+        else:
+            self.__semaphore.acquire()
 
-        self.__value = self.getValue()
+        self.__value = self.getValue(mutex=True)
 
         self.__value.__delitem__(key)
 
-        self.setValue(self.__value)
+        self.setValue(self.__value, mutex=True)
+
+        self.__semaphore.release()
 
     def __repr__(self):
         """Redefined method to print value of the Client Class instance
@@ -719,7 +770,7 @@ class SharedMemory:
         Returns:
             str: printable value of Client Class instance
         """
-        _s = "Client: " + str(self.__name) + "\n"\
+        _s = "Client: " + str(self.__name_memory) + "\n"\
             + "\tAvailable: " + self.getAvailability().__repr__() + "\n"\
             + "\tValue: " + self.__value.__repr__()
 
